@@ -10,7 +10,7 @@ pub mod stream;
 pub mod sync;
 
 use crate::error::{Result, ZipError};
-use crate::Compression;
+use crate::spec::compression::Compression;
 
 use std::convert::TryInto;
 use std::pin::Pin;
@@ -20,6 +20,7 @@ use async_compression::tokio::bufread::{BzDecoder, DeflateDecoder, LzmaDecoder,
 use chrono::{DateTime, Utc};
 use crc32fast::Hasher;
 use tokio::io::{AsyncRead, AsyncReadExt, AsyncWrite, BufReader, ReadBuf, Take};
+use async_io_utilities::{AsyncDelimiterReader, AsyncPrependReader};
 
 /// An entry within a larger ZIP file reader.
 #[derive(Debug)]
@@ -91,26 +92,70 @@ impl ZipEntry {
     pub fn compression(&self) -> &Compression {
         &self.compression
     }
+}
 
-    /// Returns the offset at which data for this entry starts.
-    pub(crate) fn data_offset(&self) -> u64 {
-        30 + self.offset.unwrap() as u64 + (self.name().len() + self.extra().unwrap().len()) as u64
+pub(crate) enum PrependReader<'a, R: AsyncRead + Unpin> {
+    Normal(OwnedReader<'a, R>),
+    Prepend(OwnedReader<'a, AsyncPrependReader<R>>),
+}
+
+impl<'a, R: AsyncRead + Unpin> AsyncRead for PrependReader<'a, R> {
+    fn poll_read(mut self: Pin<&mut Self>, c: &mut Context<'_>, b: &mut ReadBuf<'_>) -> Poll<tokio::io::Result<()>> {
+        match *self {
+            PrependReader::Normal(ref mut inner) => Pin::new(inner).poll_read(c, b),
+            PrependReader::Prepend(ref mut inner) => Pin::new(inner).poll_read(c, b)
+        }
+    }
+}
+
+pub(crate) enum OwnedReader<'a, R: AsyncRead + Unpin> {
+    Owned(R),
+    Borrow(&'a mut R),
+}
+
+impl<'a, R: AsyncRead + Unpin> AsyncRead for OwnedReader<'a, R> {
+    fn poll_read(mut self: Pin<&mut Self>, c: &mut Context<'_>, b: &mut ReadBuf<'_>) -> Poll<tokio::io::Result<()>> {
+        match *self {
+            OwnedReader::Owned(ref mut inner) => Pin::new(inner).poll_read(c, b),
+            OwnedReader::Borrow(ref mut inner) => Pin::new(inner).poll_read(c, b)
+        }
+    }
+}
+
+pub(crate) enum LocalReader<'a, R: AsyncRead + Unpin> {
+    Standard(CompressionReader<PrependReader<'a, R>>),
+    Stream(CompressionReader<AsyncDelimiterReader<PrependReader<'a, R>>>),
+}
+
+impl<'a, R: AsyncRead + Unpin> AsyncRead for LocalReader<'a, R> {
+    fn poll_read(mut self: Pin<&mut Self>, c: &mut Context<'_>, b: &mut ReadBuf<'_>) -> Poll<tokio::io::Result<()>> {
+        match *self {
+            LocalReader::Standard(ref mut inner) => Pin::new(inner).poll_read(c, b),
+            LocalReader::Stream(ref mut inner) => Pin::new(inner).poll_read(c, b)
+        }
     }
 }
 
 /// A ZIP file entry reader which may implement decompression.
 pub struct ZipEntryReader<'a, R: AsyncRead + Unpin> {
     pub(crate) entry: &'a ZipEntry,
-    pub(crate) reader: CompressionReader<'a, R>,
+    pub(crate) reader: LocalReader<'a, R>,
     pub(crate) hasher: Hasher,
     pub(crate) consumed: bool,
-    pub(crate) stream: bool,
+    pub(crate) data_descriptor: Option<(u32, u32, u32)>,
 }
 
 impl<'a, R: AsyncRead + Unpin> ZipEntryReader<'a, R> {
     /// Construct an entry reader from its raw parts (a shared reference to the entry and an inner reader).
-    pub(crate) fn from_raw(entry: &'a ZipEntry, reader: CompressionReader<'a, R>, stream: bool) -> Self {
-        ZipEntryReader { entry, reader, stream, hasher: Hasher::new(), consumed: false }
+    pub(crate) fn from_raw(entry: &'a ZipEntry, reader: CompressionReader<PrependReader<'a, R>>, _: bool) -> Self {
+        let reader = LocalReader::Standard(reader);
+        ZipEntryReader { entry, reader, hasher: Hasher::new(), consumed: false, data_descriptor: None }
+    }
+
+    /// Construct an entry reader from its raw parts (a shared reference to the entry and an inner reader).
+    pub(crate) fn with_data_descriptor(entry: &'a ZipEntry, reader: CompressionReader<AsyncDelimiterReader<PrependReader<'a, R>>>, _: bool) -> Self {
+        let reader = LocalReader::Stream(reader);
+        ZipEntryReader { entry, reader, hasher: Hasher::new(), consumed: false, data_descriptor: None }
     }
 
     /// Returns a reference to the inner entry's data.
@@ -126,7 +171,38 @@ impl<'a, R: AsyncRead + Unpin> ZipEntryReader<'a, R> {
     /// Returns true if the computed CRC32 value of all bytes read so far matches the expected value.
     pub fn compare_crc(&mut self) -> bool {
         let hasher = std::mem::take(&mut self.hasher);
-        self.entry.crc32().unwrap() == hasher.finalize()
+        let final_crc = hasher.finalize();
+
+        if self.entry().data_descriptor() {
+            self.data_descriptor.unwrap().0 == final_crc
+        } else {
+            self.entry().crc32().unwrap() == final_crc
+        }
+    }
+
+    pub(crate) async fn reset_reader(&mut self) -> Result<()> {
+        if let LocalReader::Stream(ref mut inner) = self.reader {
+            let inner_mut = inner.get_mut();
+            inner_mut.reset();
+
+            let crc = inner_mut.read_u32_le().await?;
+            let compressed = inner_mut.read_u32_le().await?;
+            let uncompressed = inner_mut.read_u32_le().await?;
+
+            self.data_descriptor = Some((crc, compressed, uncompressed));
+
+            let mut buffer = Vec::new();
+            buffer.extend_from_slice(inner_mut.buffer());
+            
+            if let PrependReader::Prepend(inner) = inner_mut.get_mut() {
+                match inner {
+                    OwnedReader::Owned(inner) => inner.prepend(&buffer),
+                    OwnedReader::Borrow(inner) => inner.prepend(&buffer),
+                };
+            }
+        }
+
+        Ok(())
     }
 
     /// A convenience method similar to `AsyncReadExt::read_to_end()` but with the final CRC32 check integrated.
@@ -136,6 +212,8 @@ impl<'a, R: AsyncRead + Unpin> ZipEntryReader<'a, R> {
         let mut buffer = Vec::with_capacity(self.entry.uncompressed_size.unwrap().try_into().unwrap());
         self.read_to_end(&mut buffer).await?;
 
+        self.reset_reader().await?;
+
         if self.compare_crc() {
             Ok(buffer)
         } else {
@@ -150,6 +228,8 @@ impl<'a, R: AsyncRead + Unpin> ZipEntryReader<'a, R> {
         let mut buffer = String::with_capacity(self.entry.uncompressed_size.unwrap().try_into().unwrap());
         self.read_to_string(&mut buffer).await?;
 
+        self.reset_reader().await?;
+
         if self.compare_crc() {
             Ok(buffer)
         } else {
@@ -170,6 +250,8 @@ impl<'a, R: AsyncRead + Unpin> ZipEntryReader<'a, R> {
         let mut reader = BufReader::with_capacity(buffer, &mut self);
         tokio::io::copy_buf(&mut reader, writer).await.unwrap();
 
+        self.reset_reader().await?;
+
         if self.compare_crc() {
             Ok(())
         } else {
@@ -197,53 +279,46 @@ impl<'a, R: AsyncRead + Unpin> AsyncRead for ZipEntryReader<'a, R> {
     }
 }
 
-impl<'a, R: AsyncRead + Unpin> Drop for ZipEntryReader<'a, R> {
-    fn drop(&mut self) {
-        if self.stream && !self.consumed {
-            panic!("Not all bytes of this reader were consumed before being dropped.");
-        }
-    }
-}
-
 /// A reader which may implement decompression over its inner type, and of which supports owned inner types or mutable
-/// borrows of them. Implements identical compression types to that of the crate::Compression enum.
+/// borrows of them. Implements identical compression types to that of the crate::spec::compression::Compression enum.
 ///
 /// This underpins entry reading functionality for all three sub-modules (stream, seek, and concurrent).
-pub(crate) enum CompressionReader<'a, R: AsyncRead + Unpin> {
+pub(crate) enum CompressionReader<R: AsyncRead + Unpin> {
     Stored(Take<R>),
-    StoredBorrow(Take<&'a mut R>),
     Deflate(DeflateDecoder<BufReader<Take<R>>>),
-    DeflateBorrow(DeflateDecoder<BufReader<Take<&'a mut R>>>),
     Bz(BzDecoder<BufReader<Take<R>>>),
-    BzBorrow(BzDecoder<BufReader<Take<&'a mut R>>>),
     Lzma(LzmaDecoder<BufReader<Take<R>>>),
-    LzmaBorrow(LzmaDecoder<BufReader<Take<&'a mut R>>>),
     Zstd(ZstdDecoder<BufReader<Take<R>>>),
-    ZstdBorrow(ZstdDecoder<BufReader<Take<&'a mut R>>>),
     Xz(XzDecoder<BufReader<Take<R>>>),
-    XzBorrow(XzDecoder<BufReader<Take<&'a mut R>>>),
 }
 
-impl<'a, R: AsyncRead + Unpin> AsyncRead for CompressionReader<'a, R> {
+impl<R: AsyncRead + Unpin> CompressionReader<R> {
+    pub(crate) fn get_mut(&mut self) -> &mut R {
+        match self {
+            CompressionReader::Stored(inner) => inner.get_mut(),
+            CompressionReader::Deflate(inner) => inner.get_mut().get_mut().get_mut(),
+            CompressionReader::Bz(inner) => inner.get_mut().get_mut().get_mut(),
+            CompressionReader::Lzma(inner) => inner.get_mut().get_mut().get_mut(),
+            CompressionReader::Zstd(inner) => inner.get_mut().get_mut().get_mut(),
+            CompressionReader::Xz(inner) => inner.get_mut().get_mut().get_mut(),
+        }
+    } 
+}
+
+impl<R: AsyncRead + Unpin> AsyncRead for CompressionReader<R> {
     fn poll_read(mut self: Pin<&mut Self>, c: &mut Context<'_>, b: &mut ReadBuf<'_>) -> Poll<tokio::io::Result<()>> {
         match *self {
             CompressionReader::Stored(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::StoredBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
             CompressionReader::Deflate(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::DeflateBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
             CompressionReader::Bz(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::BzBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
             CompressionReader::Lzma(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::LzmaBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
             CompressionReader::Zstd(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::ZstdBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
             CompressionReader::Xz(ref mut inner) => Pin::new(inner).poll_read(c, b),
-            CompressionReader::XzBorrow(ref mut inner) => Pin::new(inner).poll_read(c, b),
         }
     }
 }
 
-impl<'a, R: AsyncRead + Unpin> CompressionReader<'a, R> {
+impl<'a, R: AsyncRead + Unpin> CompressionReader<R> {
     pub(crate) fn from_reader(compression: &Compression, reader: Take<R>) -> Self {
         match compression {
             Compression::Stored => CompressionReader::Stored(reader),
@@ -254,17 +329,6 @@ impl<'a, R: AsyncRead + Unpin> CompressionReader<'a, R> {
             Compression::Xz => CompressionReader::Xz(XzDecoder::new(BufReader::new(reader))),
         }
     }
-
-    pub(crate) fn from_reader_borrow(compression: &Compression, reader: Take<&'a mut R>) -> Self {
-        match compression {
-            Compression::Stored => CompressionReader::StoredBorrow(reader),
-            Compression::Deflate => CompressionReader::DeflateBorrow(DeflateDecoder::new(BufReader::new(reader))),
-            Compression::Bz => CompressionReader::BzBorrow(BzDecoder::new(BufReader::new(reader))),
-            Compression::Lzma => CompressionReader::LzmaBorrow(LzmaDecoder::new(BufReader::new(reader))),
-            Compression::Zstd => CompressionReader::ZstdBorrow(ZstdDecoder::new(BufReader::new(reader))),
-            Compression::Xz => CompressionReader::XzBorrow(XzDecoder::new(BufReader::new(reader))),
-        }
-    }
 }
 
 macro_rules! reader_entry_impl {
@@ -283,6 +347,11 @@ macro_rules! reader_entry_impl {
             }
             None
         }
+
+        /// Returns an optional ending comment.
+        pub fn comment(&self) -> Option<&str> {
+            self.comment.as_ref().map(|x| &x[..])
+        }
     };
 }
 
