@@ -8,22 +8,24 @@
 //! ```
 
 use crate::error::{Result, ZipError};
-use crate::header::LocalFileHeader;
-use crate::read::{CompressionReader, ZipEntry, ZipEntryReader};
-use crate::Compression;
+use crate::read::{CompressionReader, ZipEntry, ZipEntryReader, OwnedReader, PrependReader};
+use crate::spec::compression::Compression;
+use crate::spec::header::LocalFileHeader;
 
 use tokio::io::{AsyncRead, AsyncReadExt};
+use async_io_utilities::{AsyncPrependReader, AsyncDelimiterReader};
 
 /// A reader which acts over a non-seekable source.
-pub struct ZipFileReader<'a, R: AsyncRead + Unpin> {
-    pub(crate) reader: &'a mut R,
+pub struct ZipFileReader<R: AsyncRead + Unpin> {
+    pub(crate) reader: AsyncPrependReader<R>,
     pub(crate) entry: Option<ZipEntry>,
     pub(crate) finished: bool,
 }
 
-impl<'a, R: AsyncRead + Unpin> ZipFileReader<'a, R> {
+impl<R: AsyncRead + Unpin> ZipFileReader<R> {
     /// Constructs a new ZIP file reader from a mutable reference to a reader.
-    pub fn new(reader: &'a mut R) -> Self {
+    pub fn new(reader: R) -> Self {
+        let reader = AsyncPrependReader::new(reader);
         ZipFileReader { reader, entry: None, finished: false }
     }
 
@@ -38,7 +40,7 @@ impl<'a, R: AsyncRead + Unpin> ZipFileReader<'a, R> {
 
         if self.finished {
             return Ok(None);
-        } else if let Some(inner) = read_lfh(self.reader).await? {
+        } else if let Some(inner) = read_lfh(&mut self.reader).await? {
             self.entry = Some(inner);
         } else {
             self.finished = true;
@@ -48,21 +50,29 @@ impl<'a, R: AsyncRead + Unpin> ZipFileReader<'a, R> {
         let entry_borrow = self.entry.as_ref().unwrap();
 
         if entry_borrow.data_descriptor() {
-            return Err(ZipError::FeatureNotSupported("Entries with data descriptors"));
+            let delimiter = crate::spec::signature::DATA_DESCRIPTOR.to_le_bytes();
+            let reader = OwnedReader::Borrow(&mut self.reader);
+            let reader = PrependReader::Prepend(reader);
+            let reader = AsyncDelimiterReader::new(reader, &delimiter);
+            let reader = CompressionReader::from_reader(entry_borrow.compression(), reader.take(u64::MAX));
+
+            Ok(Some(ZipEntryReader::with_data_descriptor(entry_borrow, reader, true)))
+        } else {
+            let reader = OwnedReader::Borrow(&mut self.reader);
+            let reader = PrependReader::Prepend(reader);
+            let reader = reader.take(entry_borrow.compressed_size.unwrap().into());
+            let reader = CompressionReader::from_reader(entry_borrow.compression(), reader);
+
+            Ok(Some(ZipEntryReader::from_raw(entry_borrow, reader, true)))
         }
-
-        let reader = self.reader.take(entry_borrow.compressed_size.unwrap().into());
-        let reader = CompressionReader::from_reader_borrow(entry_borrow.compression(), reader);
-
-        Ok(Some(ZipEntryReader::from_raw(entry_borrow, reader, true)))
     }
 }
 
 pub(crate) async fn read_lfh<R: AsyncRead + Unpin>(reader: &mut R) -> Result<Option<ZipEntry>> {
     match reader.read_u32_le().await? {
-        crate::delim::LFHD => {}
-        crate::delim::CDFHD => return Ok(None),
-        actual => return Err(ZipError::UnexpectedHeaderError(actual, crate::delim::LFHD)),
+        crate::spec::signature::LOCAL_FILE_HEADER => {}
+        crate::spec::signature::CENTRAL_DIRECTORY_FILE_HEADER => return Ok(None),
+        actual => return Err(ZipError::UnexpectedHeaderError(actual, crate::spec::signature::LOCAL_FILE_HEADER)),
     };
 
     let header = LocalFileHeader::from_reader(reader).await?;
@@ -76,7 +86,7 @@ pub(crate) async fn read_lfh<R: AsyncRead + Unpin>(reader: &mut R) -> Result<Opt
         crc32: Some(header.crc),
         uncompressed_size: Some(header.uncompressed_size),
         compressed_size: Some(header.compressed_size),
-        last_modified: crate::utils::zip_date_to_chrono(header.mod_date, header.mod_time),
+        last_modified: crate::spec::date::zip_date_to_chrono(header.mod_date, header.mod_time),
         extra: Some(extra),
         compression: Compression::from_u16(header.compression)?,
         offset: None,
