@@ -47,15 +47,14 @@
 pub(crate) mod compressed_writer;
 pub(crate) mod entry_stream;
 pub(crate) mod entry_whole;
-pub(crate) mod offset_writer;
 
 pub use entry_stream::EntryStreamWriter;
 
 use crate::error::Result;
-use crate::header::{CentralDirectoryHeader, EndOfCentralDirectoryHeader};
-use crate::Compression;
+use crate::spec::compression::Compression;
+use crate::spec::header::{CentralDirectoryHeader, EndOfCentralDirectoryHeader};
 use entry_whole::EntryWholeWriter;
-use offset_writer::OffsetAsyncWriter;
+use async_io_utilities::AsyncOffsetWriter;
 
 use tokio::io::{AsyncWrite, AsyncWriteExt};
 
@@ -95,16 +94,16 @@ pub(crate) struct CentralDirectoryEntry {
 ///
 /// # Note
 /// - [`ZipFileWriter::close()`] must be called before a stream writer goes out of scope.
-pub struct ZipFileWriter<'a, W: AsyncWrite + Unpin> {
-    pub(crate) writer: OffsetAsyncWriter<&'a mut W>,
+pub struct ZipFileWriter<W: AsyncWrite + Unpin> {
+    pub(crate) writer: AsyncOffsetWriter<W>,
     pub(crate) cd_entries: Vec<CentralDirectoryEntry>,
     comment_opt: Option<String>,
 }
 
-impl<'a, W: AsyncWrite + Unpin> ZipFileWriter<'a, W> {
+impl<W: AsyncWrite + Unpin> ZipFileWriter<W> {
     /// Construct a new ZIP file writer from a mutable reference to a writer.
-    pub fn new(writer: &'a mut W) -> Self {
-        Self { writer: OffsetAsyncWriter::from_raw(writer), cd_entries: Vec::new(), comment_opt: None }
+    pub fn new(writer: W) -> Self {
+        Self { writer: AsyncOffsetWriter::new(writer), cd_entries: Vec::new(), comment_opt: None }
     }
 
     /// Write a new ZIP entry of known size and data.
@@ -113,7 +112,7 @@ impl<'a, W: AsyncWrite + Unpin> ZipFileWriter<'a, W> {
     }
 
     /// Write an entry of unknown size and data via streaming (ie. using a data descriptor).
-    pub async fn write_entry_stream<'b>(&'b mut self, options: EntryOptions) -> Result<EntryStreamWriter<'a, 'b, W>> {
+    pub async fn write_entry_stream<'b>(&'b mut self, options: EntryOptions) -> Result<EntryStreamWriter<'b, W>> {
         EntryStreamWriter::from_raw(self, options).await
     }
 
@@ -134,7 +133,7 @@ impl<'a, W: AsyncWrite + Unpin> ZipFileWriter<'a, W> {
         let cd_offset = self.writer.offset();
 
         for entry in &self.cd_entries {
-            self.writer.write_all(&crate::delim::CDFHD.to_le_bytes()).await?;
+            self.writer.write_all(&crate::spec::signature::CENTRAL_DIRECTORY_FILE_HEADER.to_le_bytes()).await?;
             self.writer.write_all(&entry.header.to_slice()).await?;
             self.writer.write_all(entry.opts.filename.as_bytes()).await?;
             self.writer.write_all(&entry.opts.extra).await?;
@@ -151,7 +150,7 @@ impl<'a, W: AsyncWrite + Unpin> ZipFileWriter<'a, W> {
             file_comm_length: self.comment_opt.as_ref().map(|v| v.len() as u16).unwrap_or_default(),
         };
 
-        self.writer.write_all(&crate::delim::EOCDD.to_le_bytes()).await?;
+        self.writer.write_all(&crate::spec::signature::END_OF_CENTRAL_DIRECTORY.to_le_bytes()).await?;
         self.writer.write_all(&header.to_slice()).await?;
         if let Some(comment) = self.comment_opt {
             self.writer.write_all(comment.as_bytes()).await?;
